/*------------------------------------------------------------------*\
	'TabLaunch' by Dung N. Lam, based on YeahLaunch 0.2 by knorke

	Copyright (c) 2005 Dung N. Lam

	file: tablaunch.c
	author: Dung N. Lam

	$Id: tablaunch.c,v 1.39 2005/04/20 01:49:09 dnlam Exp $

	This program is free software; you can redistribute it and/or modify 
	it under the terms of the GNU General Public License as published by 
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of   
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details. 

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software 
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\*------------------------------------------------------------------*/



/*------------------------------------------------------------------*\
	headers
\*------------------------------------------------------------------*/
#include <X11/Xlib.h>
#include <X11/extensions/shape.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>

#ifdef ___IMLIB2___
	// do "apt-get install libimlib2-dev" to get header
	// defines __IMLIB_API_H
	#include <Imlib2.h>
#else
	// do "apt-get install imlib11-dev" to get header
	// defines __IMLIB_H__
	#include <Imlib.h>  
	#define Imlib_Image ImlibImage
#endif

 /* Linux with glibc:
	 *   _REENTRANT to grab thread-safe libraries
	 *   _POSIX_SOURCE to get POSIX semantics
  */
#ifdef __linux__
#  define _REENTRANT
#endif
 
 /* Hack for LinuxThreads */
#ifdef __linux__
#  define _P __P
#endif
#include <pthread.h>

/*------------------------------------------------------------------*\
	global constants
\*------------------------------------------------------------------*/
#ifndef DEBUG
	#define DEBUGOUT if(0)
#else
	#define DEBUGOUT
#endif

#define MAX_STRING_LENGTH 255
#define EOL 10

const char* PROGNAME="tablaunch";
const char* VERSION="0.6";

const char* SMALL_ICON_SUBDIR = "/48x48";
const char* BIG_ICON_SUBDIR = "/64x64";

enum tl_DockPosition {
	TOP = 1,
	BOTTOM = 2,
	LEFT = 3,
	RIGHT = 4
};

/*------------------------------------------------------------------*\
	declaration of some help-structures
\*------------------------------------------------------------------*/
enum tl_Mode {
	NORMAL = 0,
	DRAGGING = 2
};

struct tl_Colors {

	XColor fg;
	XColor fg_selected;
	XColor fg_pressed;
	XColor bg;
};

struct tl_Fonts {
	XFontStruct* font;
	XFontStruct* selected;
};

struct tl_GContexts {
	GC gc;
	GC selected;
	GC pressed;
	GC blink;
};

struct tl_Options {

	int mode;				/* mode */
	int dock_position;

	char* font;
	char* font_selected;
	char* color_fg;
	char* color_fg_selected;
	char* color_fg_pressed;
	char* color_bg;
	char* configfile;
	char* iconhome;

	int x; // x offset
	int y; // y offset
	int step;
	int blinks;

	Bool right_align;
	int height;
	float brightness;

	int margin_sides;
	int margin_top;
	int margin_bottom;

	int overlap;
	int bevel;

	Bool settings;

	Bool fit_text;
	Bool restore_selectedTabOnHide;

	Bool use_iconmask;
	Bool use_iconhome;
	Bool use_optheight;
	Bool use_text_only;

	Bool show_label;
	Bool show_always_label;
	Bool show_caption;
};

struct tl_Globals {

	Display* dpy;            /* connection to xserver */
	Window  root;            /* the rootwindow */
	int screen;              /* current screen */
	
	Window launchbar;        /* needed to draw a line when icons */
	int width;               /* width of tablaunch */
	int height;              /* height of tablaunch */
	long tabEventMask;       /* event mask for tabs or icons of tabs */ 
	
	Bool visible;
	Bool allowHideThread;
	Bool hideThreadRunning;
	Bool movingThreadRunning;
	
	/* path */
	const char* home;        /* path to home */
	char* iconhomeSmall;     /* path to small icons */
	char* iconhomeBig;       /* path to bigger icons */

	struct tl_Colors colors; /* colors */
	struct tl_Fonts fonts;   /* fonts */
	struct tl_GContexts gcs; /* graphic contexts */

#ifndef ___IMLIB2___
	ImlibData* image_data;
#endif

};

struct tl_Tab {

	struct tl_Tab* prev;
	struct tl_Tab* next;

	char* label;
	char* cmd;

	Window win;
	Window caption;

	/* tab info for restored/normal state */
	int x; /* normal x position */
	int y; /* normal y position */
	int win_width;
	int win_height;
	int text_x;
	int text_y;

	/* normal image */
	Imlib_Image* image;
	Pixmap xpm;
	Pixmap mask;

	/* expanded image */
	Imlib_Image* image_exp;
	Pixmap xpm_exp;
	Pixmap mask_exp;

	/* pressed expanded image*/
	Pixmap xpm_pressed_exp;
	Pixmap mask_pressed_exp;

	// tab info for expanded state
	Bool expanded;
	int win_expanded_width;
	int win_expanded_height;
	int expanded_x;
	int expanded_y;
	int text_expanded_x;
	int text_expanded_y; // relative within window, not absolute wrt screen

};

struct tl_Tabs {
	struct tl_Tab* head; /* all tabs in a row :) */
	struct tl_Tab* selected; /* moused-over */
	struct tl_Tab* prev_selected_tab; /* moused-over before */
};


struct tl_Globals global = {

	NULL, 0, 0, 0, 0, 0, /* dpy, root, screen, launchbar, width, height */
	EnterWindowMask|LeaveWindowMask|ButtonPressMask|ButtonReleaseMask|ExposureMask, /* tabEventMask */
	False,               /* visible */
	True,                /* allowHideThread */
	False,               /* hideThreadRunning */
	False,               /* movingThreadRunning */
	NULL
};
struct tl_Tabs tabs;
struct tl_Options opts = {
	NORMAL,              /* mode */
	TOP,                 /* top */
	"fixed",             /* font */
	"fixed",             /* font_selected */
	"#ffffff",           /* color_fg */
	"#ffff00",           /* color_fg_selected */
	"#00ff00",           /* color_fg_pressed */
	"#000099",           /* color_bg */
	".tablaunch",        /* configfile */
	NULL,                /* iconhome */
	0,                   /* x offset */
	0,                   /* y offset */
	4,                   /* step */
	3,                   /* blink */
	False,               /* right_align */
	35,                  /* height */
	0.1f,                /* brightness */
	8,                   /* margin_sides */
	3,                   /* margin_top */
	3,                   /* margin_bottom */
	2,                   /* overlap */
	2,                     /* bevel */
	False,               /* settings */
	False,               /* fit text */
	True,                /* restore_selectedTabOnHide */
	True,                /* use_iconmask */
	True,                /* use_iconhome */
	False,               /* use_optheight */
	False,               /* use_text_only */
	True,                /* show_label */
	False,               /* show_always_label */
	True,                /* show_caption */
};

// gumz: why aren't the following in Global or Options?

pthread_t hidingThread;  /* holds thread info */
pthread_t movingThread;  /* holds thread info */

int yCoeff=1;  // dependent on dock_position

int hidden_height = 1;
int limit_y = 0; //-(margin_top);
int selected_yoffset = 0; //(margin_top+margin_bottom)/3;
float expand_factor = (float)64/48;

int hideDelay = 2;
int pulldown_y;
int pulldown_height;

int pulldown_height = 15;

int y_init; // initial position

int caption_yoffset = 0;


/*------------------------------------------------------------------*\
	function-prototypes (signatures)
\*------------------------------------------------------------------*/

int main(int argc, char *argv[]);
void showSettings();
//char* tl_concatPath(const char*, const char*);
char* tl_concatPath(const char *path, const char *relfile);
Window tl_createCaption(struct tl_Tab* tab);
void tl_createNewTab(struct tl_Tab* tab);
Window tl_createWindow(struct tl_Tab* tab);
void tl_deinitialize();
void tl_drawImage(Pixmap* pixmap, Pixmap* mask, Window* win, int width, int height);
void tl_drawLabel(GC gc, struct tl_Tab* tab);
char* tl_findFile(char* filename, const char* otherpath);
char* tl_findIconFile(char *filename, char *otherpath);
struct tl_Tab* tl_findTab(Window *w);
void tl_handleButtonPress(XEvent event, struct tl_Tab* tab);
void tl_hide();
void tl_initialize();
int tl_initImlib();
struct tl_Tab* tl_initNewTab(char* label, char* cmd, char* iconfile, char *iconfile2);
void tl_maximizeTab(GC gc, struct tl_Tab* tab);
void tl_moveBar(int i);
void* tl_movingBar(void* arg);
void tl_pressTab(GC gc, struct tl_Tab* tab);
void tl_processOption(const char* opt, char* val);
int tl_readNextString(FILE* fp, char* strg, int stringlength);
void tl_readConfigFile(char* filename);
void tl_refreshTabs();
void tl_restoreTab(GC gc, struct tl_Tab* tab);
int tl_roundDiv(int numerator, int denom);
void tl_setXDefaults();
void tl_showHelp();
/*------------------------------------------------------------------*\
	helper functions
\*------------------------------------------------------------------*/
int tl_roundDiv(int numerator, int denom){
	float x=(float) numerator/denom;
	return ((x)>=0?(int)((x)+0.5):(int)((x)-0.5));
}

/** find the tab to a given window */
struct tl_Tab* tl_findTab(Window *w) {
	struct tl_Tab* tab = NULL;

	for (tab = tabs.head; tab; tab = tab->next) {
		if (*w == tab->win)
			return tab;
	}
	return NULL;
}

/** combine path + relfile, result needs to be freed to
 *  avoid memleaks */
char* tl_concatPath(const char *path, const char *relfile) {
	const size_t sLength = strlen(path) + 1 + strlen(relfile);

	/*doesn't work: char aLabel[sLength+1]; // extra char for \0 */
	char *hfilename = (char*)malloc((sLength + 1)*sizeof(char));
	strcpy(hfilename, path);
	strcat(hfilename, "/");
	strcat(hfilename, relfile);

	return hfilename;
}

/** tries to find 'filename' in current path, falls back to
 *  'otherpath' */
char* tl_findFile(char* filename, const char* otherpath) {

	if (filename==NULL)
		return NULL;

	FILE* fp = NULL;

	fp = fopen(filename, "r");
	if (fp==NULL) {
		DEBUGOUT fprintf(stderr, "! Could not open file %s. \n", filename);
		if ( otherpath && filename[0] != '/') {
	
			/* TODO: possible memleak */
			char* hfilename = tl_concatPath(otherpath, filename);

			DEBUGOUT fprintf(stderr, "Trying file %s. \n", hfilename);

			fp = fopen(hfilename, "r");
			if (fp==NULL) {
				free(hfilename);
			} else
				filename = hfilename;
		}
	}

	if (fp==NULL) {
		return NULL;
	}
	DEBUGOUT fprintf(stderr, "Success! \n");
	fclose(fp);

	return filename;
}

void tl_spawn(char *cmd)
{
	pid_t pid;

	if (!(pid = fork())) {
		setsid();
		switch (fork()) {

		case 0:
			execlp("/bin/sh", "sh", "-c", cmd, NULL);
		default:
			_exit(0);
		}
	}
	if (pid > 0)
		wait(NULL);
}


/*--------------------------------------------------------*\
	init functions
\*--------------------------------------------------------*/

/** initializes imlib */
int tl_initImlib() {
#ifdef ___IMLIB2___
	Visual* vis = NULL;
	Colormap cm;
	int depth;

	/* get default visual , colormap etc. you could ask imlib2 for what it */
	/* thinks is the best, but this example is intended to be simple */
	vis = DefaultVisual(global.dpy, global.screen);
	depth = DefaultDepth(global.dpy, global.screen);
	cm = DefaultColormap(global.dpy, global.screen);

	/* set the display , visual, colormap and drawable we are using */
	imlib_context_set_display(global.dpy);
	imlib_context_set_visual(vis);
	imlib_context_set_colormap(cm);

	/* set our cache to 2 Mb so it doesn't have to go hit the disk as long as */
	/* the images we use use less than 2Mb of RAM (that is uncompressed) */
	//   imlib_set_cache_size(2048 * 1024);
	/* set the font cache to 512Kb - again to avoid re-loading */
	//   imlib_set_font_cache_size(512 * 1024);
	/* add the ./ttfonts dir to our font path - you'll want a notepad.ttf */
	/* in that dir for the text to display */
	//   imlib_add_path_to_font_path("./ttfonts");
	/* set the maximum number of colors to allocate for 8bpp and less to 128 */
	imlib_set_color_usage(128);
	/* dither for depths < 24bpp */
	imlib_context_set_dither(1);
	// for better alpha blending
	imlib_context_set_blend(1);
#else
	/* Immediately afterwards Intitialise Imlib */
	global.image_data = Imlib_init(global.dpy);
#endif
	return 1;
}

void tl_initialize(){
	struct tl_Tab* tab = NULL;
	XSetWindowAttributes attrib;
	int bar_y;

	if (opts.dock_position == BOTTOM) {
		yCoeff = -1;
		bar_y = XHeightOfScreen(ScreenOfDisplay(global.dpy, global.screen))  - opts.y;
		pulldown_y = bar_y - pulldown_height - global.height;
		bar_y -= hidden_height;
	} else {
		bar_y = opts.y;
		yCoeff = 1;
		pulldown_y = bar_y + pulldown_height;
	}

	y_init = bar_y + yCoeff * (-global.height + hidden_height - opts.bevel);
	global.width = 0;

	for (tab = tabs.head; tab; tab = tab->next) {
		tl_createNewTab(tab);
		global.width += tab->win_width + 2 * opts.bevel - opts.overlap;
	}
	DEBUGOUT fprintf(stderr, "opts.right_align=%i\n", opts.right_align);

	attrib.override_redirect = True;
	attrib.background_pixel = global.colors.bg.pixel;

	global.launchbar = XCreateWindow(global.dpy, global.root,
			  opts.x, bar_y, global.width, hidden_height,
			  0, CopyFromParent, InputOutput, CopyFromParent, CWOverrideRedirect|CWBackPixel,
			  &attrib);
	XSelectInput(global.dpy, global.launchbar,
				 EnterWindowMask|LeaveWindowMask|ExposureMask|ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);
	XMapWindow(global.dpy, global.launchbar);

	//puts("about to move");	sleep(2);
	if (opts.right_align) {
		int screenwidth = XWidthOfScreen(ScreenOfDisplay(global.dpy, global.screen));
		int i = (screenwidth - global.width); // screenwidth - width of entire launchbar
		i-= opts.x; // opts.x is the x coordinate of where the launchbar is initially located
		i-= opts.x; // also use opts.x for offsetting the launchbar from the right side of screen
		tl_movingBar(&i); //use function directly instead of tl_moveBar(), which creates a detached thread
		opts.x += i; // update with the new x coordinate of the launchbar
	}
	//puts("moved");  sleep(2);

	// move tabs before mapping 
	for (tab = tabs.head; tab; tab = tab->next) {
		XSelectInput(global.dpy, tab->win, global.tabEventMask);
		XMapWindow(global.dpy, tab->win);
		//fprintf(stderr, "  mapping %i %s\n", tab->y, tab->label);
	}

	tl_hide();

	//puts("mapped");
	//	printf("width=%i %i\n", width, hidden_height);
	//printf("%i %i %i\n", tl_roundDiv(4,3), tl_roundDiv(1,2), tl_roundDiv(5,3));
}

void tl_deinitialize(){
	if (global.launchbar){
		XUnmapWindow(global.dpy, global.launchbar);
		XDestroyWindow(global.dpy, global.launchbar);
	}
}


// -------------------------------------------
//
void tl_showSettings(){
	printf("%s version %s\n", PROGNAME, VERSION);
	puts("---------------------");
#ifdef ___IMLIB2___
	puts("(compiled using Imlib2)");
#else
	puts("(compiled using Imlib1)");
#endif
	printf("overlap = %i\n", opts.overlap);
	printf("bevel = %i\n", opts.bevel);
	printf("pulldown_height = %i\n", pulldown_height);
	printf("margin_sides = %i\n", opts.margin_sides);
	printf("margin_top = %i\n", opts.margin_top);
	printf("margin_bottom = %i\n", opts.margin_bottom);
	printf("hidden_height = %i\n", hidden_height);
	printf("limit_y = %i\n", limit_y);
	printf("selected_yoffset = %i\n", selected_yoffset);
	printf("expand_factor = %f\n", expand_factor);
	printf("fit_text = %i\n", opts.fit_text);
	printf("use_icon_mask = %i\n", opts.use_iconmask);
	printf("keep_selected_on_hide = %i\n", opts.restore_selectedTabOnHide);
	printf("always_show_label = %i\n", opts.show_always_label);
	printf("text_only = %i\n", opts.use_text_only);
	printf("hide_delay = %i\n", hideDelay);
	printf("show_caption = %i\n", opts.show_caption);
	printf("show_label = %i\n", opts.show_label);
	printf("caption_yoffset = %i\n", caption_yoffset);
	printf("font = %s\n", opts.font);
	printf("selected_font = %s\n", opts.font_selected);
	printf("foreground = %s\n", opts.color_fg);
	printf("pressed_foreground = %s\n", opts.color_fg_pressed);
	printf("selected_foreground = %s\n", opts.color_fg_selected);
	printf("background = %s\n", opts.color_bg);
	printf("configfile = %s\n", opts.configfile);
	printf("use_iconhome = %i\n", opts.use_iconhome);
	printf("iconhome = %s\n", opts.iconhome);
	printf("xoffset = %i\n", opts.x);
	printf("yoffset = %i\n", opts.y);
	printf("step_factor = %i\n", opts.step);
	printf("blinks = %i\n", opts.blinks);
	printf("brightness = %f\n", opts.brightness);
	printf("rightalign = %i\n", opts.right_align);
	printf("same_height = %i\n", opts.height);

	puts(  "---------- for debugging --------------");
	printf("\t iconhomeSmall = %s\n", global.iconhomeSmall);
	printf("\t iconhomeBig = %s\n", global.iconhomeBig);

	printf("\t use same_height = %i\n", opts.use_optheight);
	printf("\t height = %i\n", opts.height);
	printf("\t width = %i\n", global.width);

	printf("\t yCoeff= %i\n", yCoeff);
	puts("---------------------");
}




void tl_drawImage(Pixmap* pixmap, Pixmap* mask, Window* win, int width, int height) {
	DEBUGOUT printf("%s\n", __FUNCTION__);
#ifdef ___IMLIB2___
/*		imlib_context_set_drawable(win);
		imlib_context_set_blend(0);
		imlib_context_set_mask(*mask);

		//imlib_render_image_on_drawable_at_size(0, 0, width, height);
*/
#else
		//puts("slower: reloads image");
		/* Put the Image in the window, at the window's size and apply a shape mask */
		/* if applicable, or remove one if not */
		//Imlib_apply_image(global.image_data,im, win);
#endif

	/* Put the Image pixmap in the background of the window */
	XSetWindowBackgroundPixmap(global.dpy, *win, *pixmap);
	if (opts.use_iconmask) {
		/* If there was a mask to the image, set the Image's mask to it */
		if (mask)
			XShapeCombineMask(global.dpy, *win, ShapeBounding, 0, 0, *mask, ShapeSet);
	}
	
	/* Clear the window to update the background change */
	XClearWindow(global.dpy, *win);

	/* Synchronise with the Xserver */
	// XSync(dpy,False); // XSync(..., true) removes all current events from queue
}

void tl_drawLabel(GC gc, struct tl_Tab* tab) {
	DEBUGOUT printf("%s : selected %d, expanded %d\n", tab->label, tab == tabs.selected, tab->expanded);

	if (!gc){
		gc = tab->expanded ? global.gcs.selected : global.gcs.gc;
	}

	if (tab->expanded /*&& t == selected_tab*/) {
		if (tab->image==NULL) {
			XDrawString(global.dpy, tab->win, gc,
			  tab->text_expanded_x, tab->text_expanded_y,
			  tab->label, strlen(tab->label));
		} else if (opts.show_caption) {
			XDrawString(global.dpy, tab->caption, gc,
			  opts.margin_sides, global.fonts.selected->ascent,
			  tab->label, strlen(tab->label));
		} else if (opts.show_label) {
			XDrawString(global.dpy, tab->win, gc,
			  tab->text_expanded_x, tab->text_expanded_y,
			  tab->label, strlen(tab->label));
		}
	} else {
		if ((opts.show_always_label && opts.show_label && opts.show_caption) || !tab->image) {
			// y coordinate is baseline of text; 
			XDrawString(global.dpy, tab->win, gc,
			  tab->text_x, tab->text_y,
			  tab->label, strlen(tab->label));
		}
	}
}

void tl_restoreTab(GC gc, struct tl_Tab* tab) {
	tab->expanded = False;
	//--XMapWindow(global.dpy, tab->win);
	XMoveResizeWindow(global.dpy, tab->win,
	  tab->x, pulldown_y,
	  tab->win_width , tab->win_height);
	//printf("pulldown= %i\n", pulldown_y);
	if (!opts.use_text_only) {
		if(tab->xpm) {
			tl_drawImage(&tab->xpm, &tab->mask,
			  &tab->win,
			  tab->win_width, tab->win_height);

			if (opts.show_caption)
				XUnmapWindow(global.dpy, tab->caption);
		}
	}
	//done by Expose: if (t->im == NULL) draw_label(gc, t);
}
void tl_maximizeTab(GC gc, struct tl_Tab* tab) {
	DEBUGOUT fprintf(stderr, "%s ___\n", __FUNCTION__);

	tab->expanded=1;
	XMoveResizeWindow(global.dpy, tab->win,
	  tab->expanded_x, tab->expanded_y,
	  tab->win_expanded_width, tab->win_expanded_height);
	XRaiseWindow(global.dpy, tab->win);

	if (!opts.use_text_only){
		if(tab->xpm_exp){
			tl_drawImage(&tab->xpm_exp, &tab->mask_exp,
			  &tab->win,
			  tab->win_expanded_width, tab->win_expanded_height);

			if (opts.show_caption){
				XMapWindow(global.dpy, tab->caption);
				XRaiseWindow(global.dpy, tab->caption);
			}
		}
	}

	tl_drawLabel(gc, tab); /* TODO: not always done correctly by Expose */

	DEBUGOUT fprintf(stderr, "%s ---\n", __FUNCTION__);
}

void tl_pressTab(GC gc, struct tl_Tab* tab) {
	DEBUGOUT fprintf(stderr, "%s ___ %s\n", __FUNCTION__, tab->label);

	XRaiseWindow(global.dpy, tab->win);
	if (!opts.use_text_only){
		if(tab->xpm_pressed_exp){
			tl_drawImage(&tab->xpm_pressed_exp, &tab->mask_pressed_exp,
			  &tab->win,
			  tab->win_expanded_width, tab->win_expanded_height);

			if (opts.show_caption){
				XRaiseWindow(global.dpy, tab->caption);
			}
		}
	}

	tl_drawLabel(gc, tab);  //not always done correctly by Expose

	DEBUGOUT fprintf(stderr, "%s --- %s\n", __FUNCTION__, tab->label);
}

void tl_refreshTabs(){
	struct tl_Tab* tab = NULL;
	for (tab = tabs.head; tab; tab = tab->next)
		tl_drawLabel(NULL, tab);
}

char* tl_findIconFile(char *filename, char *otherpath){
	DEBUGOUT fprintf(stderr, "%s: %s %s\n", __FUNCTION__, filename, otherpath);
	char *fn = tl_findFile(filename, opts.iconhome);
	if (fn == NULL){
		fn = tl_findFile(filename, otherpath);
	}
	return fn;
}

struct tl_Tab* tl_initNewTab(char* label, char* cmd, char* iconfile, char *iconfile2) {

	struct tl_Tab* tab = NULL;
	size_t sLength = 0;
	int h = 0;

	DEBUGOUT fprintf(stderr, "%s, %s, %s, %s, %s \n", __FUNCTION__, label, cmd, iconfile, iconfile2);

	tab = (struct tl_Tab*)malloc(sizeof(struct tl_Tab));
	if (tab) {

		/* first tab ? */
		if (tabs.head==NULL) {
			tab->next = tab->prev = NULL;
			tabs.head = tab;
		} else {
			tab->prev = tabs.prev_selected_tab;
			tab->next = NULL;
			tabs.prev_selected_tab->next = tab;
		}

		tabs.prev_selected_tab = tab; /* keep track of last created tab thru init-phase */

		tab->expanded=0;

		sLength = strlen(label);
		tab->label = (char *)malloc((sLength + 1)*sizeof(char));
		strcpy(tab->label, label);

		sLength = strlen(cmd);
		tab->cmd = (char*)malloc((sLength + 1)*sizeof(char));
		strcpy(tab->cmd, cmd);

		DEBUGOUT fprintf(stderr, "new created tl_Tab: %s, %s\n", tab->label, tab->cmd);

		tab->image = NULL;
		tab->image_exp = NULL;
		tab->caption = 0;
		tab->xpm = tab->xpm_exp = tab->xpm_pressed_exp = 0;
		tab->mask = tab->mask_exp = tab->mask_pressed_exp = 0;

		if (opts.use_text_only) {

		} else{

			if (iconfile != NULL) {
				char* foundiconfile = tl_findIconFile(iconfile, global.iconhomeSmall);
				if (foundiconfile == NULL){
					fprintf(stderr, "Unable to find small icon %s.\n", iconfile);
					iconfile = NULL;
				} else {

				/* Load the image */
#ifdef ___IMLIB2___
					tab->image = imlib_load_image(foundiconfile);
#else /* imlib1 */
 					tab->image = Imlib_load_image(global.image_data, foundiconfile);
#endif
					if (tab->image==NULL){
						fprintf(stderr, "Tab->image is NULL!! for %s", foundiconfile);
					}
					if (!opts.use_optheight) {
#ifdef ___IMLIB2___
						imlib_context_set_image(tab->image);

						h = imlib_image_get_height();
#else /* imlib1 */
						h = tab->image->rgb_height;
#endif
						if (h > global.height)
							global.height = h;
					}
				}
			}

			if ( !iconfile2 && iconfile) {
				iconfile2 = iconfile;
			}

			// now load the expanded image if specified
			if (iconfile2) {
				char* foundiconfile2 = tl_findIconFile(iconfile2, global.iconhomeBig);
				if (foundiconfile2 == NULL) {
					fprintf(stderr, "Unable to find big icon %s.  Using small icon in its place.\n", iconfile2);

					/*char *hfilename = tl_concatPath(global.iconhomeBig, iconfile);
					DEBUGOUT fprintf(stderr, "%s: Using small icon as big. %s \n",__FUNCTION__, hfilename);
					iconfile2 = tl_findFile(hfilename, NULL);
					free(hfilename);*/
				} else {
					/* Load the image */
#ifdef ___IMLIB2___
					tab->image_exp = imlib_load_image(foundiconfile2);
#else
 					tab->image_exp = Imlib_load_image(global.image_data, foundiconfile2);
#endif
				}
			}
		}
		return tab;
	}
	return NULL;
}



Window tl_createCaption(struct tl_Tab* tab) {

	XSetWindowAttributes attrib;
	int width;
	int height;
	int y;

	attrib.override_redirect = True;
	attrib.background_pixel = global.colors.bg.pixel;

	width = XTextWidth(global.fonts.selected,
						   tab->label, strlen(tab->label)) + 2 * opts.margin_sides;
	height = global.fonts.selected->ascent + global.fonts.selected->descent;

	DEBUGOUT fprintf(stderr, "caption window: %i %i %i %i\n",
				 tab->x, y_init + tab->win_expanded_height + 10,
				 width, height);

	y = tab->expanded_y;
	if (opts.dock_position == TOP)
		y+= tab->win_expanded_height + caption_yoffset;
	else if (opts.dock_position==BOTTOM)
		y-= height + caption_yoffset;
	return XCreateWindow(global.dpy, global.root,
	  tab->expanded_x + tl_roundDiv((tab->win_expanded_width - width), 2), y,
	  width, height,
	  opts.bevel,
	  CopyFromParent, InputOutput, CopyFromParent,
	  CWOverrideRedirect | CWBackPixel,
	  &attrib);
}

Window tl_createWindow(struct tl_Tab* tab) {
   
	XSetWindowAttributes attrib;

	if (opts.dock_position == TOP){
		tab->y = tl_roundDiv(global.height - tab->win_height, 2);
		tab->y += opts.y;
	} else if (opts.dock_position == BOTTOM) {
		int bottom_y = XHeightOfScreen(ScreenOfDisplay(global.dpy, global.screen)) - 1;
		tab->y = bottom_y - global.height + tl_roundDiv((global.height - tab->win_height), 2);
		tab->y -= opts.y;
	}

	attrib.override_redirect = True;
	attrib.background_pixel = global.colors.bg.pixel;
	return XCreateWindow(global.dpy, global.root,
	  tab->x, tab->y,
	  tab->win_width, tab->win_height,
	  opts.bevel, CopyFromParent, InputOutput, CopyFromParent,
	  CWOverrideRedirect | CWBackPixel,
	  &attrib);
}

void tl_createNewTab(struct tl_Tab* tab) {


	if (tab) {
		int text_width = 0;
		Imlib_Image* image_exp = NULL;
		int im_width = 0;
		int im_height = 0;

		DEBUGOUT fprintf(stderr, "%s -> %s\n", tab->label, tab->cmd);

		tab->x = (tab == tabs.head) ? opts.x : tab->prev->x + tab->prev->win_width + 2 * opts.bevel - opts.overlap;
		text_width = XTextWidth(global.fonts.font, tab->label, strlen(tab->label)) + 2 * opts.margin_sides;

		if (opts.use_text_only || !tab->image) {

			tab->win_width = text_width; //change this later?

			DEBUGOUT fprintf(stderr, "win_width %s: %i %i \n", tab->label, tab->win_width, opts.margin_sides);

			tab->win_height = opts.margin_top + 
			  global.fonts.font->ascent + 
			  global.fonts.font->descent + 
			  opts.margin_bottom;
			tab->win = tl_createWindow(tab);

		} else {

			//Imlib_Image *im=t->im;
#ifdef ___IMLIB2___
			imlib_context_set_image(tab->image);
			im_height = imlib_image_get_height();
			im_width = imlib_image_get_width();
#else
			im_height = tab->image->rgb_height;
			im_width = tab->image->rgb_width;
#endif
			
			tab->win_height = opts.use_optheight ? opts.height : im_height;

			if (opts.fit_text)
				tab->win_width = (im_width > text_width) ? im_width : text_width;
			else 
				tab->win_width = (int)((float)im_width * ( (float) tab->win_height / (float)im_height));

			DEBUGOUT fprintf(stderr, "icon %s: %i %i %i \n",
			  tab->label,
			  tab->win_height,
			  im_height,
			  tab->win_width);

			//t->win_height+= margin_bottom;
			tab->win = tl_createWindow(tab);

#ifdef ___IMLIB2___
			imlib_context_set_drawable(tab->win);

			/* TODO: imho not needed! */
			//tab->xpm = (Pixap) malloc(sizeof(Pixmap));
			//tab->mask = (Pixmap) malloc(sizeof(Pixmap));

			//needed?
			if (opts.use_iconmask)
				imlib_image_set_has_alpha(1);
			imlib_image_set_irrelevant_alpha(0);

			imlib_render_pixmaps_for_whole_image_at_size(&tab->xpm,
			  &tab->mask,
			  tab->win_width, tab->win_height);
			//imlib_free_pixmap_and_mask(t->xpm);
#else
 			/* Render the original 24-bit Image data into a pixmap of size w * h */
 			Imlib_render(global.image_data, tab->image, tab->win_width, tab->win_height);
			/* Extract the Image and mask pixmaps from the Image */
			tab->xpm = Imlib_move_image(global.image_data, tab->image);
			/* The mask will be 0 if the image has no transparency */
			if (opts.use_iconmask) {
				tab->mask = Imlib_move_mask(global.image_data, tab->image);
			}
#endif
		}

		if (!tab->image_exp) { /* for tabs with only text or one image */
			if (tab->image) {
#ifdef ___IMLIB2___
//				Imlib_Image im_copy=imlib_clone_image();
//				im2 = &im_copy;
				image_exp = tab->image;
#else
				image_exp = tab->image;
#endif
			}
			tab->win_expanded_width = tab->win_width * expand_factor;
			//DEBUGOUT printf("%f ex_width=%f\n", (expand_factor*(float)2), t->win_width );//, t->win_expanded_width);
			tab->win_expanded_height = tab->win_height * expand_factor;
			DEBUGOUT fprintf(stderr, "using normal image as expanded image\n");
		} else {

			image_exp = tab->image_exp;
#ifdef ___IMLIB2___
			imlib_context_set_image(tab->image_exp);
			im_height = imlib_image_get_height();
			im_width = imlib_image_get_width();
#else
			im_height = tab->image_exp->rgb_height;
			im_width = tab->image_exp->rgb_width;
#endif
			tab->win_expanded_height = im_height;
			tab->win_expanded_width = im_width; //(int) ((float) im_width * ( (float) t->win_expanded_height / (float) im_height));
		}

		if (opts.fit_text && (tab->win_expanded_width < text_width)) {
			tab->win_expanded_width = text_width;
			// do we want to adjust the height to keep the image's aspect ratio
		}

		if (image_exp) {
			DEBUGOUT fprintf(stderr, " ex_width=%i ex_height=%i\n",
			  tab->win_expanded_width, tab->win_expanded_height);
			
#ifdef ___IMLIB2___
			imlib_context_set_image(image_exp);
			
			//needed?
			if (opts.use_iconmask)
				imlib_image_set_has_alpha(1);
			imlib_image_set_irrelevant_alpha(0);

			/* TODO: not needed or wrong imho */
			//tab->xpm_exp =(Pixmap) malloc(sizeof(Pixmap));
			//tab->mask_exp = (Pixmap) malloc(sizeof(Pixmap));;
			imlib_render_pixmaps_for_whole_image_at_size(&tab->xpm_exp,
			  &tab->mask_exp,
			  tab->win_expanded_width,
			  tab->win_expanded_height);

			// need to clone image
			imlib_context_set_image(imlib_clone_image());

			Imlib_Color_Modifier colorMod = imlib_create_color_modifier();
			imlib_context_set_color_modifier(colorMod);
			imlib_modify_color_modifier_brightness(opts.brightness);
			imlib_apply_color_modifier();

			// so that this doesn't carry over to the next image
			imlib_reset_color_modifier(); // must be before rendering for some reason

			/* TODO: not needed imho */
			//tab->xpm_pressed_exp = (Pixmap) malloc(sizeof(Pixmap));
			//tab->mask_pressed_exp = (Pixmap) malloc(sizeof(Pixmap));
			imlib_render_pixmaps_for_whole_image_at_size(&tab->xpm_pressed_exp,
			  &tab->mask_pressed_exp,
			  tab->win_expanded_width,
			  tab->win_expanded_height);

			imlib_context_set_color_modifier(NULL);
#else
 			/* Render the original 24-bit Image data into a pixmap of size w * h */
 			Imlib_render(global.image_data,image_exp, tab->win_expanded_width, tab->win_expanded_height);
 			/* Extract the Image and mask pixmaps from the Image */
 			tab->xpm_exp = Imlib_move_image(global.image_data,image_exp);
			/* The mask will be 0 if the image has no transparency */
			if (opts.use_iconmask) {
				tab->mask_exp = Imlib_move_mask(global.image_data,image_exp);
			}

			ImlibColorModifier mod;
			Imlib_get_image_modifier(global.image_data,image_exp, &mod);
			mod.brightness=255*(1 + opts.brightness);
			Imlib_set_image_modifier(global.image_data,image_exp, &mod);
 			Imlib_render(global.image_data,image_exp, tab->win_expanded_width, tab->win_expanded_height);

			tab->xpm_pressed_exp = Imlib_move_image(global.image_data,image_exp);
			if (opts.use_iconmask) {
				tab->mask_pressed_exp = Imlib_move_mask(global.image_data,image_exp);
			}

			//reset for next image
			mod.brightness=255;
			Imlib_set_image_modifier(global.image_data,image_exp, &mod);
#endif
		}
		
		tab->text_x = opts.margin_sides + tl_roundDiv((tab->win_width - text_width), 2);

		int selected_text_width = XTextWidth(global.fonts.selected, tab->label, strlen(tab->label)) 
		  + 2 * opts.margin_sides;
		tab->text_expanded_x = opts.margin_sides + tl_roundDiv((tab->win_expanded_width - selected_text_width), 2);

		if (!tab->image) {
			tab->text_y = opts.margin_top + global.fonts.font->ascent;
			tab->text_expanded_y = tl_roundDiv(tab->win_expanded_height * opts.margin_top,
			   tab->win_height) + global.fonts.selected->ascent;
		} else {
			tab->text_y = tl_roundDiv(tab->win_height -
			  (global.fonts.font->ascent + global.fonts.font->descent), 2) + global.fonts.font->ascent;
			tab->text_expanded_y = tl_roundDiv(tab->win_expanded_height -
			  (global.fonts.selected->ascent + global.fonts.selected->descent), 2) + global.fonts.selected->ascent;
		}

		tab->expanded_x = tab->x - tl_roundDiv(tab->win_expanded_width - tab->win_width, 2);
		tab->expanded_y = pulldown_y - tl_roundDiv(tab->win_expanded_height - tab->win_height, 2);
		if (!tab->image)
			tab->expanded_y = pulldown_y - tl_roundDiv(tab->win_expanded_height - global.height, 2);
		if (tab->expanded_y < 0) {
			DEBUGOUT fprintf(stderr, "expanded_y = %i< 0\n", tab->expanded_y);
			tab->expanded_y = 0;
		}

		//initial rendering
		if (!tab->image) {
			//done by Expose: draw_label(gc, t);
		} else {
			if (opts.show_caption)
				tab->caption = tl_createCaption(tab);

			tl_drawImage(&tab->xpm, &tab->mask, &tab->win, tab->win_width, tab->win_height);
		}

		DEBUGOUT fprintf(stderr, "icon %s: %i %i \n", tab->label,
		  tab->win_expanded_height, tab->win_expanded_width);
		
	/* Synchronise with the Xserver */
	// XSync(dpy,False);

	} else
		fprintf(stderr, "malloc() failed!");
}

void* tl_movingBar(void* arg) {

	XWindowAttributes win_attr;
	struct tl_Tab* tab = NULL;
	int i = *(int*) arg;

	//printf("%p \t i= %i\n", arg, i);
	
	XRaiseWindow(global.dpy, global.launchbar);
	/* query the window's attributes. */
	XGetWindowAttributes(global.dpy, global.launchbar, &win_attr);
	XMoveWindow(global.dpy, global.launchbar,
	  win_attr.x + i,
	  win_attr.y);

	for (tab = tabs.head; tab; tab = tab->next) {
		tab->x += i;
		tab->expanded_x += i;
//usleep(2000);

		DEBUGOUT fprintf(stderr, " moving window %s\n", tab->label);
		XGetWindowAttributes(global.dpy, tab->win, &win_attr);
		if (tab->expanded)
			XMoveWindow(global.dpy, tab->win, tab->expanded_x, win_attr.y);
		else
			XMoveWindow(global.dpy, tab->win, tab->x, win_attr.y);

		if (tab->caption) {
			XGetWindowAttributes(global.dpy, tab->caption, &win_attr);
			XMoveWindow(global.dpy, tab->caption, win_attr.x + i, win_attr.y);
		}
	}
	//puts("stopping ...");
	tl_refreshTabs();
	XSync(global.dpy, False);
	global.movingThreadRunning = False;

	return NULL;
}

void tl_moveBar(int i) {

	int errcode;								/* holds pthread error code */
	int* deltaX = NULL;

	if (global.movingThreadRunning)
		return;

	global.movingThreadRunning = True;

	deltaX = (int*) malloc(sizeof(int));
	*deltaX = i;
	errcode = pthread_create(&movingThread, /* thread struct			 */
	  NULL,		  /* default thread attributes */
	  tl_movingBar,  /* start routine			 */
	  deltaX);	   /* arg to routine			*/
	if (errcode)
		fprintf(stderr, "pthread_create ERROR");

	errcode = pthread_detach(movingThread);
	if (errcode)
		fprintf(stderr, "pthread_detach ERROR");
}

/******** this is the thread code */
void* tl_hiding(void* arg){
	global.hideThreadRunning = True;
	//int myid=*(int *) arg;
	int j = 0;
	for(; global.allowHideThread && j < hideDelay; j++) {
		DEBUGOUT fprintf(stderr, "sleeping %i\n", j);
		sleep(1);
	}

	if(global.allowHideThread) {

		struct tl_Tab* tab = NULL;

		if (opts.restore_selectedTabOnHide && tabs.prev_selected_tab){
			// ignore to EnterNotify events in case big icon has transparent holes 
			// and small (restored) icon causes EnterNotify event
			XSelectInput(global.dpy, tabs.prev_selected_tab->win, 0); 
			tl_restoreTab(global.gcs.gc, tabs.prev_selected_tab);
		}

		for (tab = tabs.head; tab; tab = tab->next) {
			//printf("Hiding %s : %i \n", t->label, height);
			//--XMoveWindow(global.dpy, tab->win, tab->x, y_init);
			// XLowerWindow(dpy, t->win);
			XUnmapWindow(global.dpy, tab->win);
		}
		if (tabs.prev_selected_tab){
			// restore event listening
			XSelectInput(global.dpy, tabs.prev_selected_tab->win, global.tabEventMask);
		}
		XFlush(global.dpy);
		global.visible = False;
	} else {
		DEBUGOUT fprintf(stderr, "Hiding cancelled\n");
	}
	global.hideThreadRunning = False;
	pthread_exit(0);
	return 0;
}

void tl_hide(){
	if ( !global.allowHideThread || global.hideThreadRunning || opts.mode == DRAGGING)
		return;

	if ( pthread_create(&hidingThread,/* thread struct */
	  NULL,         /* default thread attributes */
	  tl_hiding,    /* start routine */
	  0)            /* arg to routine */
	){
		fprintf(stderr, "pthread_create ERROR");
	} else {
		pthread_detach(hidingThread);
	}
}

void tl_handleButtonPress(XEvent event, struct tl_Tab* tab) {

	int in = 1;
	//XDrawString(dpy, t->win, gc_selected, 3, height - 4, t->label, strlen(t->label) );
	//draw_label(gc_selected, t);

	//restore_tab(gc, t); // simulates pushed image
	tl_pressTab(global.gcs.pressed, tab);

	do { // watch for mouse dragging
		XMaskEvent(global.dpy, ButtonReleaseMask|LeaveWindowMask|EnterWindowMask, &event);
		if (event.type == LeaveNotify) {
			tl_maximizeTab(global.gcs.selected, tab);
			in = 0;
		} else if (event.type == EnterNotify && tl_findTab(&event.xcrossing.window) == tab) {
			tl_pressTab(global.gcs.pressed, tab);
			in = 1;
		}
	} while (event.type != ButtonRelease);

	// need to clear area to remove label drawn from appearing
	/* Clear the window to update the background change */
	//XClearWindow(dpy,t->win);
	
	//printf("in=%i\n", in);
	tl_maximizeTab(global.gcs.selected, tab);

	if (in) {

		tl_spawn(tab->cmd);

		if (event.xbutton.button == Button3)
			tl_hide();

		for (in = 0; in < opts.blinks; in++) {
			usleep(100000);
			tl_pressTab(global.gcs.pressed, tab);
			tl_drawLabel(global.gcs.pressed, tab);
			XSync(global.dpy, False);
			usleep(100000);
			tl_maximizeTab(global.gcs.selected, tab);
			tl_drawLabel(global.gcs.selected, tab);
			XSync(global.dpy, False);
		}
	}
}

// returns 1 if stringlength too short to fit next string
// returns 0 if successful
// returns EOF if EOF
// returns 10 if \n
int tl_readNextString(FILE* fp, char* strg, int stringlength) {
	char c = 0;
	int i = 0;
	int maxindexFor0 = stringlength - 2; // need to reserve last char for \0
	int inQuotes = 0;
	strg[0]='\0'; //strcpy(strg,""); // clear out the strg
	
	do{ // go to the next word or return EOF
		//if (c) { putchar(c); printf("skip: %i\n", c);}
		c = fgetc(fp);
		if (c == EOF) return EOF;
		if (c == '\n') return EOL;
	}while (isspace(c));

	do{
		if (i >= maxindexFor0) {
			strg[i] = '\0';
			fprintf(stderr, "String array too small!");
			return 1;
		}
		if (c == '"') {
			inQuotes = 1;
		} else {
			if (c == '\\')
				c = fgetc(fp);

			strg[i++] = c;
			//putchar(c); printf("char: %i\n", c);
		}
		c = fgetc(fp);
		if (! inQuotes){
			if (isspace(c) ||  c=='\n')
				break;
		} else if (c == '"') { //found end quote
			break;
		}

	} while ( c != EOF );

	strg[i] = '\0';
	//printf("String: %i %s\n", strlen(strg), strg);
	if (c == EOF)
		return EOF;
	else if (c == '\n')
		return EOL;
	return 0;
}

static char *help_message[] = {
	"where options include:", 
	"  (Note: X resource should be prepended with 'tablaunch.')",
	"  descriptions refer default 'top' docking position",
	"  option               X resource             description",
	"  ------               ----------             -----------",
	"  -h, --help                                  displays this help",
	"  --settings                                  lists current settings",
	"                                              (a good way to see default values)",
	"  --ignoreXRM                                 ignore X resources",
	"  -c configfile        configfile             config file listing tabs",
	"  -fg color            foreground             foreground text color", 
	"  -sfg color           selected_foreground    foreground text color of selected tab", 
	"  -pfg color           pressed_foreground     foreground text color of pressed tab", 
	"  -bg color            background             background color of tab", 
	"  -fn fontname         font                   name of font for tab", 
	"  -sfn fontname        selected_font          name of font for selected tab", 
	"  -x xoffset           xoffset                x offset of tabs from left/right of screen",
	"  -y yoffset           yoffset                y offset of tabs from top/bottom of screen",
	"  -r                   rightalign             align tabs to right of screen",
	"  -d hideDelay         hide_delay             seconds until tabs auto-hide",
	"  -l                   always_show_label      always show text label",
	"  -t                   text_only              don't use icons",
	"  -k                   keep_selected_on_hide  selected tab is apparent when hidden",
	"  -e float             expand_factor          expand factor",
	"  --iconMaskOff        use_icon_mask          allows images to be transparent",
	"  --fit_text           fit_text               makes icons wider to fit text",
	"  --s_yoffset pixels   selected_yoffset       y offset of selected tab from adjacent tabs",
	"  --limit_y pixels     limit_y                y coordinate of highest tab",
	"                                              (negative value means above screen)",
	"  --hidden pixels      hidden_height          pixels that are visible when auto-hidden",
	"  --bottom pixels      margin_bottom          space below text label",
	"  --top pixels         margin_top             space above text label",
	"  --sides pixels       margin_sides           extra space left and right of text label",
	"  --pulldown pixels    pulldown_height        how far down to pulldown the selected tab",
	"  --opts.bevel pixels  opts.bevel             size of bevel around tab",
	"  --overlap pixels     tab_overlap            pixels that tab overlap each other",
	"                                              (negative value means tab separation) ",
	"  --step pixels        step_factor            step factor of tab offset as it gets",
	"                                              farther from the selected tab",
	"  --height pixels      same_height            use this height for all icons",
	//"  --max_height pixels  max_height             use this height for enlarged icons",
	"  --caption_y pixels   caption_yoffset        y offset for captions under icons",
	"  --no_caption         no_caption             turn off caption under icons",
	"                         or show_caption      ",
	"  --no_label           no_label               turn off all labels for icons",
	"                         or show_label        ",
	"  --iconhome dir       iconhome               set dir to search for icons",
	"                                              expects 48x48 and 64x64 subdirs",
	"  --no_iconhome        no_iconhome            don't use icon home",
	"                         or use_iconhome      ",
	"  --blinks int         blinks                 number of blinks when tab is pressed",
	"  --brightness float   brightness             brightness of pressed icon",
	"  --dock [top|bottom]  dock_position          screen position of launchbar",
	NULL
};

void tl_showHelp(){
	char **cpp;
	fprintf(stdout, "\n%s version %s \n", PROGNAME, VERSION);
	fprintf (stdout, "usage:  %s [option] [option] ...\n", PROGNAME);
	for (cpp = help_message; *cpp; cpp++) {
		fprintf (stdout, "%s\n", *cpp);
	}
	fprintf (stdout, "\n");
	exit(0);
}


/*------------------------------------------------------------------*\
	parse options and commandline
\*------------------------------------------------------------------*/
void tl_processOption(const char* opt, char* val) {

	if (opt[0]=='\0')
		return;

	DEBUGOUT fprintf(stderr, "option: %s \n", opt);
	if (!strcmp(opt, "-h") || !strcmp(opt, "--help")) {
		tl_showHelp();
	} else if (!strcmp(opt, "--ignoreXRM")) {
		// taken care of already
	} else if ( !strcmp(opt, "--no_caption") || !strcmp(opt, "no_caption") ){
		if ( !val || !strcasecmp(val, "true") )
			opts.show_caption = False;
	} else if ( !strcmp(opt, "show_caption") ){
		if ( !strcasecmp(val, "false") )
			opts.show_caption=False;
	} else if ( !strcmp(opt, "--no_label") || !strcmp(opt, "no_label") ){
		if ( !val || !strcasecmp(val, "true") )
			opts.show_label=False;
	} else if ( !strcmp(opt, "show_label") ){
		if ( !strcasecmp(val, "false") )
			opts.show_label=False;
	} else if ( !strcmp(opt, "--no_iconhome") || !strcmp(opt, "no_iconhome") ){
		if ( !val || !strcasecmp(val, "true") )
			opts.use_iconhome = False;
	} else if ( !strcmp(opt, "use_iconhome") ){
		if ( !strcasecmp(val, "false") )
			opts.use_iconhome = False;
	} else if ( !strcmp(opt, "-r") || !strcmp(opt, "rightalign") ){
		if ( !val || !strcasecmp(val, "true") ) 
			opts.right_align=True;
		//printf("screen width=%d \n", opt_rx);
	} else if (!strcmp(opt, "-t") || !strcmp(opt, "text_only")){
		opts.use_text_only = True;
	} else if (!strcmp(opt, "-l") || !strcmp(opt, "always_show_label")){
		opts.show_always_label = True;
	} else if (!strcmp(opt, "-k") || !strcmp(opt, "keep_selected_on_hide")){
		opts.restore_selectedTabOnHide = 0;
	} else if (!strcmp(opt, "--iconMaskOff") || !strcmp(opt, "use_icon_mask")){
		opts.use_iconmask = False;
	} else if (!strcmp(opt, "--fit_text") || !strcmp(opt, "fit_text")){
		opts.fit_text=True;
	} else if (!strcmp(opt, "--settings") || !strcmp(opt, "settings")){
		opts.settings = True;
	} else if (val!=0){
		DEBUGOUT printf("	 =%s\n", val);
		if (!strcmp(opt, "-fn") || !strcmp(opt, "font")){
			opts.font = val;
		} else if (!strcmp(opt, "-sfn") || !strcmp(opt, "selected_font")){
			opts.font_selected = val;
		} else if (!strcmp(opt, "-fg") || !strcmp(opt, "foreground")){
			opts.color_fg = val;
		} else if (!strcmp(opt, "-sfg") || !strcmp(opt, "selected_foreground")){
			opts.color_fg_selected = val;
		} else if (!strcmp(opt, "-pfg") || !strcmp(opt, "pressed_foreground")){
			opts.color_fg_pressed = val;
		} else if (!strcmp(opt, "-c") || !strcmp(opt, "configfile")){
			opts.configfile = val;
		} else if (!strcmp(opt, "-bg") || !strcmp(opt, "background")){
			opts.color_bg = val;
		} else if (!strcmp(opt, "-x") || !strcmp(opt, "xoffset")){
			opts.x = atoi(val);
		} else if (!strcmp(opt, "-y") || !strcmp(opt, "yoffset")){
			opts.y = atoi(val);
		} else if (!strcmp(opt, "-d") || !strcmp(opt, "hide_delay")){
			hideDelay = atoi(val);
		} else if (!strcmp(opt, "-e") || !strcmp(opt, "expand_factor")){
			expand_factor = atof(val);
		} else if (!strcmp(opt, "--s_yoffset") || !strcmp(opt, "selected_yoffset")){
			selected_yoffset = atoi(val);
		} else if (!strcmp(opt, "--limit_y") || !strcmp(opt, "limit_y")){
			limit_y = atoi(val);
		} else if (!strcmp(opt, "--hidden") || !strcmp(opt, "hidden_height")){
			hidden_height = atoi(val);
		} else if (!strcmp(opt, "--bottom") || !strcmp(opt, "margin_bottom")){
			opts.margin_bottom = atoi(val);
		} else if (!strcmp(opt, "--top") || !strcmp(opt, "margin_top")){
			opts.margin_top = atoi(val);
		} else if (!strcmp(opt, "--sides") || !strcmp(opt, "margin_sides")){
			opts.margin_sides = atoi(val);
		} else if (!strcmp(opt, "--pulldown") || !strcmp(opt, "pulldown_height")){
			pulldown_height = atoi(val);
		} else if (!strcmp(opt, "--opts.bevel") || !strcmp(opt, "opts.bevel")){
			opts.bevel = atoi(val);
		} else if (!strcmp(opt, "--overlap") || !strcmp(opt, "tab_overlap")){
			opts.overlap = atoi(val);
		} else if (!strcmp(opt, "--blinks") || !strcmp(opt, "blinks")){
			opts.blinks = atoi(val);
		} else if (!strcmp(opt, "--step") || !strcmp(opt, "step_factor")){
			opts.step = atoi(val);
		} else if (!strcmp(opt, "--height") || !strcmp(opt, "same_height")){
			opts.use_optheight = True;
			opts.height = atoi(val);
		} else if (!strcmp(opt, "--caption_y") || !strcmp(opt, "caption_yoffset")){
			opts.show_caption = True;
			caption_yoffset = atoi(val);
		} else if (!strcmp(opt, "--iconhome") || !strcmp(opt, "iconhome")){
			opts.iconhome = val;
			opts.use_iconhome = True;
		} else if (!strcmp(opt, "--brightness") || !strcmp(opt, "brightness")){
			opts.brightness = atof(val);
		} else if (!strcmp(opt, "--dock") || !strcmp(opt, "dock_position")){
			if (!strcasecmp(val, "top"))
				opts.dock_position=TOP;
			else if (!strcasecmp(val, "bottom"))
				opts.dock_position=BOTTOM;
		} else if (!strcmp(opt, "-p") 
				|| !strcmp(opt, "propertyfile") // temporary: for backward compatibility
				){
			opts.configfile = val;
		} else if (!strcmp(opt, "--top_limit") 
				|| !strcmp(opt, "top_limit") // temporary: for backward compatibility
				){
			limit_y=atoi(val);
		} else {
			printf("Unknown option or invalid value: %s=%s\n", opt, val);
		}
	} else {
		printf("Unknown option or missing value: %s\n", opt);
	}
	
}

static char *XDEFAULT_OPTIONS[]={
	"configfile", 
	"foreground", 
	"selected_foreground", 
	"pressed_foreground", 
	"background",
	"font", 
	"selected_font", 
	"xoffset", 
	"rxoffset", 
	"rightalign", 
	"hide_delay", 
	"text_only", 
	"always_show_label",
	"keep_selected_on_hide", 
	"use_icon_mask", 
	"fit_text", 
	"expand_factor", 
	"selected_yoffset", 
	"limit_y",
	"hidden_height", 
	"margin_top", "margin_bottom", "margin_sides", 
	"pulldown_height", 
	"bevel",
	"tab_overlap", 
	"step_factor", 
	"blinks", 
	"same_height", 
	"caption_yoffset", 
	"show_caption", 
	"no_caption",
	"show_label", 
	"no_label", 
	"use_iconhome", 
	"no_iconhome", 
	"iconhome", 
	"brightness", 
	"dock_position",
 // temp: for backward compatibility
	"propertyfile", "top_limit"
};

void tl_setXDefaults(){
	char* defStr = XGetDefault(global.dpy, "tablaunch", "enable");

	char** optionKey;
	for(optionKey = XDEFAULT_OPTIONS; *optionKey; optionKey++) {
		if ( (defStr = XGetDefault(global.dpy, "tablaunch", *optionKey)) != NULL)
			tl_processOption(*optionKey, defStr);
	}
}

void tl_readConfigFile(char* filename) {
	filename = tl_findFile(filename, global.home);
	FILE* fp = fopen(filename, "r");
	if (!fp){
		fprintf(stderr, "Could not find config file! Exiting.\n");
		exit(1);
	}
	// printf("file %s\n", filename);

	char label[MAX_STRING_LENGTH];
	char cmd[MAX_STRING_LENGTH];
	char icon[MAX_STRING_LENGTH];
	char icon2[MAX_STRING_LENGTH];

	int retval;
	do {
		retval = tl_readNextString(fp, label, 255);
		if (label[0]=='#'){
			// empty line or comment
			char c=0;
			//printf("---------------- %c=%i  %i, %i \n", c, c, EOF, '\n');
			do {
				c = fgetc(fp);
				//printf("Skipping comment: %c=%i  %i, %i \n", c, c, EOF, '\n');
			} while ( (c != '\n') && (c != EOF) );

			label[0]='\0';
			continue;
		} else {
			if (retval != EOL && retval!=EOF) {
				retval = tl_readNextString(fp, cmd, 255);
				// cmd may still be empty
				if (retval != EOL && retval!=EOF) {
					retval = tl_readNextString(fp, icon, 255);
					if (retval != EOL && retval!=EOF) {
						retval = tl_readNextString(fp, icon2, 255);
					} else {
						icon2[0]='\0';
					}
				} else {
					icon[0]='\0';
					icon2[0]='\0';
				}
			} else {
				cmd[0]='\0';
				icon[0]='\0';
				icon2[0]='\0';
			}
		}
		
		// printf("label[0]=%c\n", label[0]);
		if (label[0]=='\0') {
			// blank line or comment
		} else if (label[0]=='!') {
			if (icon[0]=='\0')
				tl_processOption(cmd, 0);
			else
				tl_processOption(cmd, icon);
		} else if (cmd[0]=='\0'){
			fprintf(stderr, "ERROR: Must have command for %s", label);
		} else if (icon[0]=='\0'){
			//printf("No icon for %s\n", label);
			struct tl_Tab* newTab = tl_initNewTab(label, cmd, NULL, NULL);
			DEBUGOUT fprintf(stderr, "Created %s\n\n", newTab->label);
		} else if (icon2[0]=='\0') {
			struct tl_Tab* newTab = tl_initNewTab(label, cmd, icon, NULL);
			DEBUGOUT fprintf(stderr, "Created %s\n\n", newTab->label);
		} else {
			struct tl_Tab* newTab = tl_initNewTab(label, cmd, icon, icon2);
			DEBUGOUT fprintf(stderr, "Created %s\n\n", newTab->label);
		}
	} while (retval != EOF);
	
	fclose(fp);
}



int main(int argc, char *argv[]) {

	int i = 0;
	XEvent event;
	XGCValues gv;
	struct tl_Tab* tab = NULL;
	Bool ignoreXRM = False;
	size_t sLength;
	XColor dummy;

	//if (argc == 1) showHelp();
	i = XInitThreads(); // must be the first Xlib function a multi-threaded program calls; needed under XFree
	if (i==0) {
		fprintf(stderr, "! XInitThreads() unsuccessful: %i \n", i);
		exit(1);
	}

	/* set some global variables */
	global.home = getenv("HOME");
	if (!(global.dpy = XOpenDisplay(NULL))) {
		fprintf(stderr, " can't open dpy %s", XDisplayName(NULL));
		exit(1);
	}
	global.screen = DefaultScreen(global.dpy);
	global.root = RootWindow(global.dpy, global.screen);

	/* init imlib - stuff */
	if (!tl_initImlib()) {
		fprintf(stderr, "! couldnt init imlib.\n");
		exit(1);
	}

	/* use .Xdefaults ? */
	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "--ignoreXRM")) {
			ignoreXRM = True;
			break;
	}
	}

	if (! ignoreXRM)
		tl_setXDefaults();

	opts.dock_position=TOP;

	/* parse options */
	for (i = 1; i < argc; i++) {
		// printf("arg[%i]=%s \n",i,argv[i]);
		if ( (i+1)<argc && argv[i+1][0]=='-')
			tl_processOption(argv[i], 0);
		else {
			tl_processOption(argv[i], argv[i+1]);
			++i;
		}
	}
	
	//printf("Done with options %i \n", NULL);
	XAllocNamedColor(global.dpy, DefaultColormap(global.dpy, global.screen),
	  opts.color_fg, &global.colors.fg, &dummy);
	XAllocNamedColor(global.dpy, DefaultColormap(global.dpy, global.screen),
	  opts.color_fg_pressed, &global.colors.fg_pressed, &dummy);
	XAllocNamedColor(global.dpy, DefaultColormap(global.dpy, global.screen),
	  opts.color_fg_selected, &global.colors.fg_selected, &dummy);
	XAllocNamedColor(global.dpy, DefaultColormap(global.dpy, global.screen),
	  opts.color_bg, &global.colors.bg, &dummy);

	global.fonts.font = XLoadQueryFont(global.dpy, opts.font);
	if (! global.fonts.font) {
		fprintf(stderr, "Could not load font: %s. Try 'fixed' or '*-fixed-*' instead.", opts.font);
		exit(1);
	}

	global.height = opts.use_optheight ? opts.height :
		global.fonts.font->ascent +
		global.fonts.font->descent + 
		opts.margin_top + opts.margin_bottom;

	if (opts.use_iconhome) {
		if (opts.iconhome == NULL) {
			sLength = strlen(global.home) + strlen("/.icons");
			opts.iconhome = (char*)malloc((sLength+1)*sizeof(char));
			strcpy(opts.iconhome, global.home);
			strcat(opts.iconhome, "/.icons");
		}

		sLength = strlen(opts.iconhome) + strlen(SMALL_ICON_SUBDIR);
		global.iconhomeSmall = (char *)malloc((sLength + 1)*sizeof(char));
		strcpy(global.iconhomeSmall, opts.iconhome);
		strcat(global.iconhomeSmall, SMALL_ICON_SUBDIR);

		sLength = strlen(opts.iconhome)+strlen(BIG_ICON_SUBDIR);
		global.iconhomeBig = (char *)malloc((sLength+1)*sizeof(char));
		strcpy(global.iconhomeBig, opts.iconhome);
		strcat(global.iconhomeBig, BIG_ICON_SUBDIR);
	} else {
		opts.iconhome=NULL;
		global.iconhomeSmall=NULL;
		global.iconhomeBig=NULL;
	}

	gv.font = global.fonts.font->fid;
	gv.foreground = global.colors.fg.pixel;
	gv.function = GXcopy;
	
	global.gcs.gc = XCreateGC(global.dpy, global.root, GCFunction|GCForeground|GCFont, &gv);
	global.fonts.selected = XLoadQueryFont(global.dpy, opts.font_selected);
	if (! global.fonts.selected) {
		fprintf(stderr,	"Could not load font: %s. Try 'fixed' or '*-fixed-*' instead.", opts.font_selected);
		exit(1);
	}
 //selected_height = font-> /* max_bounds. */ ascent + font-> /* max_bounds. */ descent + margin_top;

	gv.font = global.fonts.selected->fid;
	gv.foreground = global.colors.fg_selected.pixel;
	global.gcs.selected = XCreateGC(global.dpy, global.root, GCFunction|GCForeground|GCFont, &gv);

	gv.foreground = global.colors.fg_pressed.pixel;
	global.gcs.pressed = XCreateGC(global.dpy, global.root, GCFunction|GCForeground|GCFont, &gv);

	//XAllocNamedColor(dpy, DefaultColormap(dpy, screen), "#00FF00", &fg, &dummy);
	gv.foreground = global.colors.bg.pixel;
	global.gcs.blink = XCreateGC(global.dpy, global.root, GCFunction|GCForeground|GCFont, &gv);

	tl_readConfigFile(opts.configfile); //after height is initialized

	if (opts.settings)
		tl_showSettings();

	if (! tabs.head) {
		fprintf(stderr, "No tabs created!  Edit the config file");
		exit(1);
	}

	tl_initialize();
	global.visible=True;

	int y = pulldown_height - selected_yoffset;

	int buttonX = 0; // x position for button press
	Time lastTime = 0;

	XSync(global.dpy,False); 
	while (1) {
		XNextEvent(global.dpy, &event); // waits and returns when event occurs
		DEBUGOUT fprintf(stderr, "  Event=%i\n", event.type);

		switch (event.type) {
		case EnterNotify:
			DEBUGOUT fprintf(stderr, "enter \n");

			global.allowHideThread=False;
			if (global.hideThreadRunning) {
				// not yet: hideThreadRunning=0;
				i=pthread_cancel(hidingThread);
				DEBUGOUT fprintf(stderr, "  cancelled thread %p %i\n", &hidingThread, i);
				global.hideThreadRunning=False;
			}

			tabs.selected = tl_findTab(&event.xcrossing.window);
			if (tabs.prev_selected_tab && tabs.selected != tabs.prev_selected_tab) { // previously selected tab
				tl_restoreTab(global.gcs.gc, tabs.prev_selected_tab);
				//prev_selected_tab=NULL;
			}

			if (tabs.selected) {
				DEBUGOUT fprintf(stderr, "enter	  %s\n", tabs.selected->label);
		
				i = 0;
				int dy = y;
				for (tab = tabs.selected->next; tab; tab = tab->next) {

					if (dy > opts.step * i)
						dy = yCoeff * tl_roundDiv(y, (++i * opts.step));
					else if(dy)
						dy = 0;
					XMapWindow(global.dpy, tab->win);
					XMoveWindow(global.dpy, tab->win, tab->x, tab->y + dy);
				}

				i = 0;
				dy = y;
				for (tab = tabs.selected->prev; tab; tab = tab->prev) {
					if (dy > opts.step * i)
						dy = yCoeff * tl_roundDiv(y, (++i * opts.step));
					else
						if(dy) dy = 0;
					XMapWindow(global.dpy, tab->win);
					XMoveWindow(global.dpy, tab->win, tab->x, tab->y + dy);
				}
				// Expose will take care of it:	   draw_label(gc, t);
			} else {
				// if none selected
				if (!global.visible){
					//puts("---------------------- \n");
					for (tab = tabs.head; tab; tab = tab->next) {
						// move windows so they can be seen
						XMapWindow(global.dpy, tab->win);
						XMoveWindow(global.dpy, tab->win, tab->x, tab->y);
					}
				}
			}

		 if (!global.visible) {
				////printf("raising all , %s\n", head_tab->label);
				
				tab = tabs.head;
				if (tab != tabs.selected)
					for (; tab && tab != tabs.selected; tab = tab->next) {
						XRaiseWindow(global.dpy, tab->win);
				}

				while (tab && tab->next)
					tab = tab->next; // go to the end

				// printf("enter 3.6\n");
				if (tab != tabs.selected)
					for (; tab && tab != tabs.selected; tab = tab->prev) {
						XRaiseWindow(global.dpy, tab->win);
				}
				global.visible=True;
			}

			// printf("enter 4\n");
			if (tabs.selected)
				tl_maximizeTab(global.gcs.selected, tabs.selected);

			//printf("enter done %s\n", selected_tab->label);
			DEBUGOUT fprintf(stderr, "enter done \n");
			break;

		case LeaveNotify:
			DEBUGOUT if (tabs.selected) fprintf(stderr, "leaving %s\n", tabs.selected->label);
					//find_tab(event.xcrossing.window)->label);

			global.allowHideThread=True;

			tabs.prev_selected_tab = tabs.selected;
			tabs.selected=NULL;

			// causes enter-leave cycle if using icon mask with icons with holes
			//if (prev_selected_tab != NULL) restore_tab(gc, prev_selected_tab);
/*
			if (event.xcrossing.y >= pulldown_height * expand_factor
				|| event.xcrossing.x_root <= opts.x
				|| event.xcrossing.x_root >= opts.x + global.width){
				}
*/
			tl_hide();
			//printf("leave done %s\n", find_tab(event.xcrossing.window)->label);
			break;
		case Expose:
			//break;
			if (event.xexpose.count == 0) {
				tab = tl_findTab(&event.xexpose.window);
				//printf("Expose %p\n", tab);
				DEBUGOUT if (tab) fprintf(stderr, "Expose %s \n", tab->label);
				if (tab) {
					if (opts.show_always_label || !tab->image)
						tl_drawLabel(NULL, tab);
					else {
						if (tab->expanded)
							tl_drawLabel(NULL, tab);
					}
				}
				//needed? XFlush(dpy);
			}
			break;
		case ButtonRelease:
			opts.mode = NORMAL;
			//refresh_tabs();
			break;
		case ButtonPress:
			if (event.xbutton.window == global.launchbar){
				opts.mode = DRAGGING;
				lastTime = event.xmotion.time;
				buttonX = event.xmotion.x_root;
			} else {
				tabs.selected = tl_findTab(&event.xbutton.window);
				if (tabs.selected) {
					if (event.xbutton.button == Button1) {
						DEBUGOUT fprintf(stderr, "Button1 pressed\n");
						tl_handleButtonPress(event, tabs.selected);
					} else if (event.xbutton.button == Button3){
						fprintf(stderr, "Button3 pressed\n");
					} else if (event.xbutton.button == Button2){
						fprintf(stderr, "Button2 pressed\n");
					} else if (event.xbutton.button == Button4){
						DEBUGOUT fprintf(stderr, "Button4 pressed\n");
						//moveBar(-scrollIncrement);
						//XGrabPointer(dpy, selected_tab->win, False, ExposureMask, 0,0,selected_tab->win);
					} else if (event.xbutton.button == Button5){
						DEBUGOUT fprintf(stderr, "Button5 pressed\n");
						//moveBar(scrollIncrement);
					}
				}
			}
			break;
		case MotionNotify: // dragging
			DEBUGOUT printf("Bar: %i %i -> %i %i\n", event.xmotion.x_root, event.xmotion.y_root, event.xmotion.x, event.xmotion.y);
			/* if the 1st mouse button was held during this event, */
			//if (event.xmotion.state & Button1Mask) { // this is always true as long as mask it correctly set
			if (opts.mode == DRAGGING
			  && !global.movingThreadRunning
			  && XQLength(global.dpy) < 2000
			  && (event.xmotion.time-lastTime) > 0 
			) {
				tl_moveBar(event.xmotion.x_root-buttonX);
				DEBUGOUT printf("\t\t\tmove %i\n", event.xmotion.x_root-buttonX);
				lastTime = event.xmotion.time;
				buttonX = event.xmotion.x_root;
			} else{
				DEBUGOUT fprintf(stderr, "skipping %ld \n", /*(unsigned long)*/ event.xmotion.time-lastTime);
			}
			XSync(global.dpy, True); //removes all current events from queue.  in case mouse moves too much and events pile up
			break;

   case SelectionClear:
			fprintf(stderr, "got a SelectionClear event\n");
	  /* If this application previously called
	   * XSetSelectionOwner, it may get this event;
	   * otherwise, you don't need it in your
	   * event loop */
	  break;
   case SelectionNotify:
			fprintf(stderr, "got a SelectionNotify event\n");
	  /* If this application calls XConvertSelection,
	   * it will get this event; otherwise, you don't
	   * need it in your event loop */
	  break;
   case SelectionRequest:
			fprintf(stderr, "got a SelectionRequest event\n");
	  /* If this application previously called
	   * XSetSelectionOwner, it may get this event;
	   * otherwise, you don't need it in your
	   * event loop */
	  break;
		
		}
	}
	return 0;
}


